--------------------------------------------------------------------------------
-- This is a class to draw the texture. <br>
-- Base Classes => DisplayObject, TextureDrawable, Resizable <br>
--------------------------------------------------------------------------------

-- import
local table                     = require "hp/lang/table"
local class                     = require "hp/lang/class"
local Group                     = require "hp/display/Group"
local math                      = require "hp/lang/math"
local MOAIPropUtil              = require "hp/util/MOAIPropUtil"
local Component                 = require "hp/gui/Component"

-- class
local super                     = Group
local M                         = class(Group, MOAIPropUtil)
local MOAIPropInterface         = MOAIProp.getInterfaceTable()

require "global_variables"

M.TOP_LEFT = SCREEN_CORNER.TOP_LEFT
M.TOP_RIGHT = SCREEN_CORNER.TOP_RIGHT
M.BOT_LEFT = SCREEN_CORNER.BOT_LEFT
M.BOT_RIGHT = SCREEN_CORNER.BOT_RIGHT

--------------------------------------------------------------------------------
-- The constructor.
-- @param params (option)Parameter is set to Object.<br>
--------------------------------------------------------------------------------
function M:init(params)
    super.init(self, params)
    local border_size = 3
    local total_width = params.size[1] + border_size * 2
    local total_height = params.size[2] + border_size * 2
    self:setSize(total_width, total_height)
    
    local borderGraphic = Graphics { 
      size = { total_width, total_height },
      parent = self,
      pos = { 0, 0 }
    }
  
    borderGraphic:setPenColor(0.8, 0.8, 0.8, 1.0)
                  :fillRect(0 , 0, self:getWidth(), self:getHeight())
                  
    mainTexture = Sprite({  texture = params.texture, 
                            layer = params.layer, 
                            parent = self,
                            left = border_size, 
                            top = border_size, 
                            width = params.size[1],
                            height = params.size[2]})
        
    print("TangoPIP init() ")
    self._touching = false
    self._touchIndex = nil
    self:addEventListener(Component.EVENT_TOUCH_DOWN, self.touchDownHandler, self)
    self:addEventListener(Component.EVENT_TOUCH_MOVE, self.touchMoveHandler, self)
    self:addEventListener(Component.EVENT_TOUCH_CANCEL, self.touchCancelHandler, self)
    self:addEventListener(Component.EVENT_TOUCH_UP, self.touchUpHandler, self)
    self:setMargin(10,10,10,10)
    
    self.dragEndedCB = params.dragEnded
    self.dragBeganCB = params.dragBegan
    self.shouldMoveToNewCorner = params.shouldMoveToNewCorner
    self.currentCorner = self.BOT_LEFT
end

function M:setMargin(left, top, right, bottom)
  self._margin = { 
                    left = left,
                    top = top, 
                    right = right, 
                    bottom = bottom 
                  }
end



function M:touchDownHandler(e)  
  
  if(self.dragBeganCB) then
    self.dragBeganCB()
  end
  
  if self.action then
      self.action:stop()
  end
  self.touching = true
  self.action = self:seekScl(1.1, 1.1, 1, 0.5)
end

function M:touchMoveHandler(e)
  if self.touching then
      local scale = self:getLayer():getViewScale()
      local moveX, moveY = e.moveX / scale, e.moveY / scale
      -- print("moveX:" .. moveX .. " moveY:" .. moveY)      
      self:addLoc(moveX, moveY, 0)
      self._lastDelta = { x = moveX, y = moveY }
      self._lastTime  = MOAISim.getElapsedTime ()
  end
end

function M:touchUpHandler(e)
  if self.touching then
      self.action:stop()
      
      local left, top = self:getPos()      
      local corner = self:getCornerForVelocity(self._lastDelta, { x = left , y = top })
      
      if self.shouldMoveToNewCorner(corner) then
        print("may move to new corner")
      else
        print("can not to new corner, move back to old corner")
        corner = self.currentCorner  
      end  

      if(self.dragEndedCB) then
        self.dragEndedCB(corner)
      end
      
      self:animateToCorner(corner)        
      self.touching = false      
  end
end

function M:animateToCorner(corner)
  print("animateToCorner() " .. corner)
  local left, top = self:getPos()  
  local cornerPoint = self:getCornerPoint(corner)
  local scale = self:getLayer():getViewScale()      
  local targetX, targetY = cornerPoint[1] / scale, cornerPoint[2] / scale   
  local deltaTime = MOAISim.getElapsedTime () - self._lastTime
     
  
  -- calculate animation time
  local deltaPixel = math.distance(left, top, targetX, targetY)
  local velocityVector = math.distance(0, 0, self._lastDelta.x, self._lastDelta.y) / deltaTime
  local time = math.abs(deltaPixel / velocityVector)
  print("delta " .. deltaPixel .. " velocity " .. velocityVector .. " animation time  " .. time)
  self.action = Animation({self}, math.min(0.6, time), MOAIEaseType.LINEAR):moveLoc(targetX - left, targetY - top, 0):seekScl(1, 1, 1, 0.10)
  self.action:play()
  self.currentCorner = corner
end  

function M:moveToCorner(corner)
  print("moveToCorner() " .. corner)
  self:setPos(table.unpack(self:getCornerPoint(corner)))
  self.currentCorner = corner
end  

function M:getCornerForVelocity(velocity, location)
  local tts;
  local outMargin = 50;
  local pw, ph = self:getLayer():getSize()
  local newX , newY = location.x, location.y
  local div = math.max(math.abs(velocity.x) , math.abs(velocity.y))

  if div == 0 then
    --return location
  end

  while(newX >= -outMargin and newX < pw and newY >= 0 and newY <= ph) do
    newX = newX + (velocity.x / div);
    newY = newY + (velocity.y / div);
    
    ---print("## newx " .. newX .. " newY " .. newY)
  end
  
  -- print("# oldx " .. location.x .. " y " .. location.y)  
  -- print("# velocity " .. velocity.x .. " y " .. velocity.y)
  -- print("# newx " .. newX .. " newY " .. newY)
  local corner = self:getClosestCorner(newX, newY);
  return corner;
end

function M:getClosestCorner(x, y)
  local pw, ph = self:getLayer():getSize()
  local midX = pw/2;
  local midY = ph/2;
  
  if(x < midX) then
    if(y < midY) then
      return self.TOP_LEFT;
    else
      return self.BOT_LEFT;
    end
  else
    if(y < midY) then
      return self.TOP_RIGHT;
    else
      return self.BOT_RIGHT;
    end
  end
end

function M:getCornerPoint(corner)
  local pw, ph = self:getLayer():getSize()
  if(corner == self.TOP_LEFT) then
    return  { self._margin.left, self._margin.top }
  end
  if(corner == self.TOP_RIGHT) then
    return  { pw - self._margin.right - self:getWidth(), self._margin.top }
  end
  if(corner == self.BOT_LEFT) then
    return  { self._margin.left, ph - self._margin.bottom - self:getHeight() }
  end
  if(corner == self.BOT_RIGHT) then
    return  { pw - self._margin.right - self:getWidth(), ph - self._margin.bottom - self:getHeight()}
  end  
end

function M:touchCancelHandler(e)
  self:touchUpHandler(e)
end

return M